r"""
Chain maps between ground field resolutions

AUTHORS: - Christian Nassau (2009-05-02: version 1.0)

This page explains how to compute chain maps between ground field resolutions.
Applications of this are:

   - ring structure of `{\rm Ext_A({\mathbb F}_p,{\mathbb F}_p)}`
   - change of algebra maps `{\rm Ext_A} \rightarrow {\rm Ext_B}` for `B\subset A`
   - zero'th Steenrod operation `P^0:{\rm Ext}^{s,t}\rightarrow {\rm Ext}^{s,pt}`

INTRODUCTION:

Assume two Steenrod algebra ground field resolutions `C_\ast` and `D_\ast`.
Given a cocycle `g:C_s\rightarrow{\mathbb F}_p` one can ask for a chain map
`\phi:C_{s+k}\rightarrow D_k` covering `g`. In Sage such a map is represented
by an object of class ``SteenrodAlgebraChainMap``::

    sage: # define some convenient aliases
    sage: SGFR = SteenrodAlgebraGroundFieldResolution
    sage: ChainMap = SteenrodAlgebraChainMap
    sage: # compute a resolution
    sage: p=3; smax=20; nmax=10*p^2
    sage: C = SGFR(SteenrodAlgebra(p),filename=":memory:")
    sage: C.extend(s=smax,n=nmax,quiet=True)
    sage: # pick b0 = <h0,h0,...,h0>
    sage: b0 = C.g(2,2*p^2-2*p-2)
    sage: # compute a chain map lifting b0
    sage: phi = ChainMap(C,C,basecycle=b0,filename=":memory:")
    sage: phi.extend(s=smax,n=nmax,quiet=True)
   
Using `\phi` we can pull back cycles `x:D_k\rightarrow {\mathbb F}_p` 
to `x\circ\phi:C_{s+k}\rightarrow {\mathbb F}_p`::

    sage: # pull back of the unit is the class that we started with:
    sage: phi.pullback(C.g(0,0))
    g(2,10)
    sage: # compute some powers of b0
    sage: b02 = phi.pullback(b0) ; b02     # doctest note: random sign
    g(4,20)
    sage: b03 = phi.pullback(b02) ; b03    # doctest note: random sign
    g(6,30)
    sage: # compute some other products
    sage: h0 = C.g(1,2*p-3); h1 = C.g(1,2*p^2-2*p-1); h0h1 = C.g(2,2*p^2-2*p-1)
    sage: (h0,h1,h0h1)
    (g(1,3), g(1,11), g(2,11))
    sage: b0h0 = phi.pullback(h0); b0h1 = phi.pullback(h1); b0h0h1 = phi.pullback(h0h1)
    sage: (b0h0,b0h1)                      # doctest note: random sign
    (2*g(3,13), 2*g(3,21))
    sage: b0h0h1
    0

We can compute the induced map `{\rm Ext}_A\rightarrow {\rm Ext_B}` for `B\subset A`
from a chain map that covers the unit cycle::

    sage: # compute resolutions for Ext_A and Ext_A(2) at the prime 2
    sage: Ares = SGFR(SteenrodAlgebra(2),filename=":memory:")
    sage: A2res = SGFR(prime=2,profile="0 {3 2 1}",viewtype="even",filename=":memory:")
    sage: smax = 20; nmax = 40
    sage: Ares.extend(s=smax,n=nmax,quiet=True)
    sage: A2res.extend(s=smax,n=nmax,quiet=True)
    sage: # compute a comparison map A2res -> Ares. topologically, this
    sage: # corresponds to the Hurewicz map \pi_*(S) -> tmf_*
    sage: hmp = ChainMap(A2res,Ares,basecycle=A2res.g(0,0),filename=":memory:")
    sage: hmp.extend(s=smax,n=nmax)
    sage: # compute some pullbacks
    sage: def print_pullbacks(genlist):
    ...        for g in genlist:
    ...            print "%s -> %s" % (g,hmp.pullback(g))
    ...        
    sage: print_pullbacks(Ares.generators(n=24))   # doctest note: random ordering 
    g(5,24) -> 0
    g(10,24) -> 0
    g(11,24) -> g(11,24,1)
    sage: # a little check to illustrate that the pullback is additive
    sage: a = Ares.g(9,23); b = Ares.g(9,23,1)
    sage: print_pullbacks([a,b,a-3*b])               # doctest note: random ordering
    g(9,23) -> 0
    g(9,23,1) -> g(9,23)
    g(9,23) + g(9,23,1) -> g(9,23)

A ``ChainMap`` can also be used to compute the Steenrod operation 
`P^0:{\rm Ext}^{s,t} \rightarrow {\rm Ext}^{s,pt}`. This map is induced
by the algebra map `A\rightarrow A` dual to the Frobenius `A_\ast\rightarrow A_\ast`::

    sage: # compute a resolution for A(1)
    sage: A1res = SGFR(prime=2,profile="0 {2 1}",viewtype="even",filename=":memory:")
    sage: A1res.extend(s=smax,n=nmax,quiet=True)
    sage: # compute the "doubling map" A2res -> A1res
    sage: frob = ChainMap(A2res,A1res,algmap="frobenius",filename=":memory:")
    sage: frob.extend(s=smax,n=nmax,quiet=True)
    sage: # pull back A(1)'s periodicity generator
    sage: frob.pullback(A1res.g(4,8))
    g(4,20)

Finally, you can open an SQL-console to directly investigate the underlying SQLite database::

    sage: # open an SQL console
    sage: frob.sqlcon()  # sage doctest note: not tested


CLASS DOCUMENTATION:
"""

#*****************************************************************************
#       Copyright (C) 2009 Christian Nassau <nassau@nullhomotopie.de>
#  Distributed under the terms of the GNU General Public License (GPL)
#*****************************************************************************

from string import atoi
from sage.structure.sage_object import SageObject
from sage.sets.set import Set
from sage.algebras.stgfr import SteenrodAlgebraGroundFieldResolution, Yacop
from sage.rings.all import GF
from sage.structure.formal_sum import FormalSum, FormalSums

class SteenrodAlgebraChainMap(SageObject):
    """
    A chain map between ground field resolutions.
    """

    def __init__(self,src,dst,basecycle=None,algmap='inclusion',filename=":memory:"):

        import sage

        if not isinstance(src,SteenrodAlgebraGroundFieldResolution):
            raise ValueError, "first argument is not a resolution"
        if not isinstance(dst,SteenrodAlgebraGroundFieldResolution):
            raise ValueError, "second argument is not a resolution"

        if src._prime != dst._prime:
            raise ValueError, "prime mismatch between source and destination"

        if not algmap is None:
            if algmap != 'inclusion' and algmap != 'frobenius':
                raise ValueError, "algmap must be 'inclusion' or 'frobenius'"
            if algmap == 'frobenius':
                if not basecycle is None:
                    raise ValueError, "basecycle not supported for Frobenius map"
                basecycle = dst.g(0,0,0)
                
        if algmap is None and basecycle is None:
            raise ValueError, "neither algmap nor basecycle given"

        self._src = src
        self._dst = dst
        self._prime = dst._prime
        self._parentGF = GF(self._prime)
        self._fsums = FormalSums(self._parentGF)
        self.tcl = Yacop.Slave()

        if not basecycle is None:
            bc = ""
            for (cf,g) in basecycle:
                bc = bc + " %d %d" % (cf,g["id"])
            basecycle = bc

        self._srcip = src.tcl.eval("namespace current")
        self._dstip = dst.tcl.eval("namespace current")
        self._filename = filename
        self.tcl.eval("""
           yacop::chainmap create chmap {%s::resolution} {%s::resolution} {%s} {%s} {%s}
           """ % (self._srcip, self._dstip, algmap or 'default', basecycle or 'default', self._filename))
        
        

    def _repr_(self):
        return "ChainMap (%s) -> (%s), filename='%s'" % (self._src, self._dst, self._filename)
    
    def _latex_(self):
        return "\\text{%s}" % self._repr_()


    def extend(self,s,n,quiet=True):
        """
        Extend the chain map to the given bidegree `(s,n)`.
        Here `s` is the homological degree and `n` the topological 
        dimension.
        """

        if quiet:
            self.tcl.eval("yacop::sectionizer quiet on")
        else:
            self.tcl.eval("yacop::sectionizer quiet off")

        self.tcl.eval("chmap extend {smax %d nmax %d}" % (s,n))


    def pullback(self,cycle):
        """
        Compute the pullback of a cycle from the target resolution.
        """
        import sage
        bc = ""
        for (cf,g) in cycle:
            bc = bc + " %d %d" % (cf,g["id"])

        res = self.tcl.eval("""
            set res {}
            foreach {cf id} [chmap pullback {%s}] {
               lappend res ($cf,$id)
            }
            return "\[[join $res ,]\]"
          """ % bc)
        #print res
        r=[(cf,self._src.g(id=gid)) for (gid,cf) in eval(res)]
        return sage.algebras.stgfr.FormalSumOfDict(r,parent=self._fsums)
            
    def sqlcon(self):
        """
        Open a SQL console to inspect the underlying database.
        """
        self.tcl.loadTk()
        self.tcl.eval("sqlconsole new [chmap db]")
        


if False:
    Yacop.tcl.eval("package forget yacop")
    execfile("local/lib/yacop/stgfr.py")
    SGFR=SteenrodAlgebraGroundFieldResolution
    ChainMap=SteenrodAlgebraChainMap
    p=3; fname = None
    C=SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(p),filename=fname)
    smax=33
    nmax=p*p*p*p*p
    C.extend(s=smax,n=nmax,quiet=False)
    x= C.g(2,10)
    mp = ChainMap(C,C,basecycle=x,algmap='inclusion',filename=":memory:")
    mp.extend(s=5,n=nmax,quiet=False)

    for g in [C.g(0,0),x]:
        print g, " -> ", mp.pullback(g)

if False:

    p=3

    execfile("local/lib/yacop/stgfr.py")

    fname = None
    #fname = ":memory:"

    #C=SGFR(SteenrodAlgebra(p),filename=fname)
    #C=SGFR(prime=2,profile='0 {3 2 1}',filename=":memory:")

    SteenrodAlgebraGroundFieldResolution=SGFR
    C=SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(p),filename=fname)

    smax=33
    nmax=p*p*p*p*p
    C.extend(s=smax,n=nmax,quiet=False)
    print len(C.generators())

    x= C.g(2,7)
    mp = ChainMap(C,C,basecycle=x,algmap='inclusion',filename=":memory:")
    print [(lambda z: (z,x[z]))(k) for k in ("s","i","e","id")]  
    #mp.extend(s=smax,n=nmax,quiet=False)

    if false:
        for g in C.generators():
            print g
            m  = ChainMap(C,C,basecycle=g,algmap='inclusion',filename=":memory:")
            m.extend(s=smax,n=nmax,quiet=False)

    D=SteenrodAlgebraGroundFieldResolution(prime=p,profile='0 full',filename=fname)
    D.extend(s=smax,n=nmax,quiet=False)
    frob = ChainMap(C,C,algmap='frobenius',filename=":memory:")
    frob.extend(s=smax,n=nmax,quiet=False)
