r"""
Resolutions of small modules

AUTHORS: - Christian Nassau (2009-04-06: version 1.0)

This page explains how to use a ground field resolution 
to compute Ext of other small modules.

INTRODUCTION:

Let `M` and `N` be left modules over Steenrod algebra.
Given a resolution `C_\ast` of `{\mathbb F_p}` we can
form the complex `M\land C_\ast` where `\land` denotes 
the tensor product `\otimes_{\mathbb F_p}` with the
diagonal action. This complex is a resolution of `M`
and can be used to compute derived functors.
In particular, we can think of `{\rm Ext}_A(M,N)`
as the homology of

.. math::
     {\rm Hom}_A(M\land C_\ast,N) = {\rm Hom_A}(N^\ast\land M\land C_\ast,{\mathbb F_p})

If `M` and `N` are small modules this can be used to
compute `{\rm Ext}_A(M,N)` very quickly if `C_\ast` is 
already available.

Let `X=N^\ast\land M` and let `\otimes_{\mathbb F_p}^R` denote a tensor product with 
action on the right factor only. The computation uses the isomorphism

.. math::
     X\otimes_{\mathbb F_p}^R C_\ast \owns x\otimes ag \mapsto \sum \pm (a'x)\land (a''g) \in X\land C_\ast 

which exhibits the freeness of `X\land C_\ast`. Up to sign the inverse is given by

.. math::
     X\land C_\ast \owns x\land ag \mapsto \sum \pm (\chi(a')x)\otimes (a''g) \in X\otimes_{\mathbb F_p}^R C_\ast

We can give `X` a right action of `A` via `xa = (-1)^{\vert x\vert \vert a\vert} \chi(a)x`
to write this more succinctly as `x\land ag \mapsto \sum \pm (xa')\otimes (a''g)`. 
This is the form that is internally used in Sage.

THE MODULE:

A smash product of the form `X\land C_\ast` is represented by
an object of class ``SteenrodAlgebraSmashResolution``.
The `X` can be any object that implements the ``SteenrodAlgebraSmashable`` interface.
An example is given by the mod `p` Moore spectrum as follows::

    sage: # set some parameters
    sage: p=5; smax=10; nmax= 3*p*p*p
    sage: # create X
    sage: X = SteenrodAlgebraSmashableExampleMoore(p)
    sage: # test the "SteenrodAlgebraSmashable" interface
    sage: X.bbox()  # doctest note: random key order
    {'emax': 1, 'emin': 0, 'imax': 1, 'imin': 0, 'smax': 0, 'smin': 0}
    sage: basis = X.basis(); basis
    {'b', 't'}
    sage: # differentials are always zero:
    sage: [X.diff(g) for g in basis]
    [0, 0]
    sage: # the action is as expected
    sage: A = SteenrodAlgebra(p)
    sage: ops = [A.P(0), A.Q(0)]
    sage: [[(op,g,X.actR(g,op)) for g in basis] for op in ops]
    [[(P(0), 'b', b), (P(0), 't', t)], [(Q_0, 'b', t), (Q_0, 't', 0)]]

Elements of the module are represented as formal sums::

    sage: type( X.actR(basis[0],A.Q(0)) )
    <class 'sage.structure.formal_sum.FormalSum'>
    sage: X.actR(basis[0],A.Q(0)).parent()
    Finite Field of size 5
    
THE RESOLUTION:

To compute `{\rm Ext}_A(M,{\mathbb F_p})` for the Moore space module we first need
a ground field resolution::

    sage: C = SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(p),filename=":memory:")
    sage: C.extend(s=smax,n=nmax,quiet=True)

We can then define and compute the smash product `X\land C_\ast`::

    sage: prod = SteenrodAlgebraSmashResolution(C,X,filename=":memory:")
    sage: prod.extend(s=smax,n=nmax,quiet=True)

Let `R_\ast = X\land C_\ast`. The object ``prod`` now knows about both 
`{\mathbb F}_p \otimes_A R_\ast` and its homology `{\rm Tor}^A(M,{\mathbb F}_p)`::

    sage: # list some bases of the smash product and its homology
    sage: for (s,n) in [(0,0),(0,1),(1,0),(1,1),(1,7),(1,8),(1,39),(2,39),(3,39),(4,39),(5,39)]:
    ...       print "(%2d,%2d): %33s %15s" % (s,n,sorted(prod.smash_basis(s,n)),prod.homology(s,n))
    ...    
    ( 0, 0):                   [('b', g(0,0))]        {g(0,0)}
    ( 0, 1):                   [('t', g(0,0))]              {}
    ( 1, 0):                   [('b', g(1,0))]              {}
    ( 1, 1):                   [('t', g(1,0))]              {}
    ( 1, 7):                   [('b', g(1,7))]        {g(1,7)}
    ( 1, 8):                   [('t', g(1,7))]        {g(1,8)}
    ( 1,39):                  [('b', g(1,39))]       {g(1,39)}
    ( 2,39):  [('b', g(2,39)), ('t', g(2,38))]              {}
    ( 3,39):  [('b', g(3,39)), ('t', g(3,38))]              {}
    ( 4,39):  [('b', g(4,39)), ('t', g(4,38))]              {}
    ( 5,39):  [('b', g(5,39)), ('t', g(5,38))]       {g(5,39)}
    sage: # find a representative of g(5,39): 
    sage: prod.cycle(prod.g(5,39))
    ('t', g(5,38))
    sage: # look at some cycles and boundaries
    sage: prod.cycles(s=4,n=39), prod.boundaries(s=4,n=39) # doctest note: random signs
    ([('t', g(4,38))], [2*('t', g(4,38))])
    sage: prod.cycles(s=2,n=39), prod.boundaries(s=2,n=39) # doctest note: random signs
    ([('t', g(2,38))], [3*('t', g(2,38))])

GRAPHICAL USER INTERFACE:

You can inspect the smash product and its homology through a GUI window::

    sage: prod.gui()  # doctest: not tested

You can also just open an SQL console to work with the underlying database::

    sage: prod.sqlcon()   # doctest: not tested 


CLASS DOCUMENTATION:
"""

#*****************************************************************************
#       Copyright (C) 2009 Christian Nassau <nassau@nullhomotopie.de>
#  Distributed under the terms of the GNU General Public License (GPL)
#*****************************************************************************

from sage.structure.sage_object import SageObject
from sage.sets.set import Set
from sage.rings.all import GF
from sage.rings.infinity import Infinity
from sage.structure.formal_sum import FormalSum, FormalSums
from sage.algebras.stgfr import Yacop, SteenrodAlgebraGroundFieldResolution

class SteenrodAlgebraSmashable(SageObject):
    """
    An abstract base class for derivable functors.
    """

    def bbox(self):
        """
        Return a bounding box for `G(A)`.

        The method should return a dictionary with the following keys::

           :smin, smax:   bounds for homological degree *s*
           :imin, imax:   bounds for internal degree *i*
           :emin, emax:   bounds for exterior algebra degree *e*

        """
        raise NotImplementedError, "bbox method not implemented"

    def basis(self,s,i,e):
        """
        Return a basis of the totalization `G(A)` in degree `(s,i,e)`.

        INPUTS::

            :self:          the derivable functor `G` 
            :s:             homological degreee
            :i:             internal degree (not to be confused with topological dimension)
            :e:             exterior algebra degree

        The basis should consist of hashable Python objects and must be iterable.
        """
        raise NotImplementedError, "basis method not implemented"

    def diff(self,baselem): 
        """
        Return the internal differential of a basis element.

        The differential should be a ``FormalSum`` with coefficient ring `GF(p)`.
        If *basel* is from *self.basis(s,i,e)* then 
        *self.diff(basel)* must be from the span of *self.basis(s-1,i,e)*.
        """
        raise NotImplementedError, "diff method not implemented"

    def action(self,baselem,a):
        """
        Return the right action of a Steenrod operation on a basis element.

        The return value should be a ``FormalSum`` with coefficient ring `GF(p)`.
        If *basel* is from *self.basis(C,s,i,e)* then 
        *self.action(basel,a)* must be from the span of *self.basis(C,s,i+p,e+q)*
        where *p* (resp. *q*) is the internal (resp. exterior algebra) degree of *a*.

        """
        raise NotImplementedError, "action method not implemented"



class SteenrodAlgebraSmashResolution(SageObject):
    """
    Object that represents `M\\otimes_A C_\\ast` for a 
    minimal resolution `C_\\ast` and a differential right `A`-module `M`.
    """

    def __init__(self,resolution,module,filename=":memory:"):

        if not isinstance(resolution,SteenrodAlgebraGroundFieldResolution):
            raise ValueError, "first argument is not a resolution"

        self._resolution = resolution
        self._module = module
        self._filename = filename
        self._prime = self._resolution._prime
        self._parentGF = GF(self._resolution._prime)
        self._fsums = FormalSums(self._parentGF)
        self.tcl = Yacop.Slave()

        # link the resolution into our Tcl interpreter
        self._resip = resolution.tcl.eval("namespace current")
        self.tcl.eval("""
           yacop::smashres create smashprod {%s::resolution} {%s} [namespace current]::sagemod {%s}
           """ % (self._resip, self._module, self._filename))
        
    def _repr_(self):
        return "SmashResolution (%s) * (%s), filename='%s'" % (self._module, self._resolution, self._filename)
    
    def _latex_(self):
        return "\\text{%s}" % self._repr_()


    def extend(self,s,n,quiet=True,smin=0,nmin=0):

        if quiet:
            self.tcl.eval("yacop::sectionizer quiet on")
        else:
            self.tcl.eval("yacop::sectionizer quiet off")

        # dump module basis and actions into database
        self._makeSageModule({"smax":s+1,"imax":2*n+s+1})

        # compute the smash product
        self.tcl.eval("smashprod compute {smin %d nmin %d smax %d nmax %d}" % (smin,nmin,s+1,n))

    def gui(self):
        """
        Open a graphical user interface. This allows you to inspect the resolution, create
        postscript charts and to run SQL commands.
        """

        self.tcl.loadTk()
        self.tcl.eval("""
              set chv [yacop::chartgui [smashprod db] {%s}]
              trace add variable [$chv forever] write "[list $chv destroy];#"
           """ % self._filename)

    def sqlcon(self):
        """
        Open a SQL console to inspect the underlying database.
        """
        self.tcl.loadTk()
        self.tcl.eval("sqlconsole new [smashprod db]")


    def _search_condition(self,s=None,n=None,i=None,e=None,nov=None):
        """
        Create a where clause from the input parameters
        """
        
        if self._resolution._viewtype == 'even':
            nexp = "(s.ideg/2-s.sdeg)"
        else:
            nexp = "(s.ideg-s.sdeg)"

        cond = ""
        if not s is None: cond += " and s.sdeg = %d" % s
        if not i is None: cond += " and s.ideg = %d" % i
        if not n is None: cond += " and %s = %d" % (nexp,n)
        if not e is None: cond += " and s.edeg = %d" % e
        if not nov is None: cond += " and (s.sdeg-s.edeg) = %d" % nov
        if cond != "": cond = "where " + cond[4:]
        return (cond,nexp)

    def homology(self,s=None,n=None,i=None,e=None,nov=None):
        """
        Return basis of the homology in a range of degrees
        """

        (cond,nexp) = self._search_condition(s=s,n=n,i=i,e=e,nov=nov)

        res = "[" + self.tcl.eval( """
               join [smashprod db eval {
                   select pydict('id',rowid,'s',s.sdeg,'i',s.ideg,'e',s.edeg,'n',%s,'num',s.basid,'enum',s.ebasid) from homology_generators as s %s
               }] ,
            """ % (nexp,cond) ) + "]"
        #print res
        import sage
        #return Set([sage.algebras.stgfr.FormalSumOfDict(self,x) for x in eval(res)])
        return Set([sage.algebras.stgfr.FormalSumOfDict([(1,sage.algebras.stgfr.Generator(self,x))],parent=self._fsums) for x in eval(res)])

    def smash_basis(self,s=None,n=None,i=None,e=None,nov=None):
        """
        Return A-basis of the smash product in a range of degrees
        """

        (cond,nexp) = self._search_condition(s=s,n=n,i=i,e=e,nov=nov)

        res = "[" + self.tcl.eval( """
               join [smashprod db eval {
                   select pydict('modgen',''''||m.sagename||'''','resgen',s.resgen) 
                   from smash_generators as s
                   join module_generators as m on m.rowid = s.modgen %s
               }] ,
            """ % (cond) ) + "]"
        return Set([(self._module._make_element_(x["modgen"]),self._resolution.g(id=x["resgen"])) for x in eval(res)])
        
    def _unpack_vector(self,row,basis):
        """
        Create a cycle from coefficient vector and basis
        """

        res = []
        idx=-1
        for coeff in row:
            idx=idx+1
            if coeff != 0:
                x = basis[idx]
                res.append((coeff,(self._module._make_element_(x["modgen"]),self._resolution.g(id=x["resgen"]))))
                
        return FormalSum(res,parent=self._fsums)

    def _getmatrix(self,what,s=None,n=None,i=None,e=None,nov=None):
        """
        Return cycles in a range of degrees
        """
        
        (cond,nexp) = self._search_condition(s=s,n=n,i=i,e=e,nov=nov)

        res = "[" + self.tcl.eval( """
               join [smashprod db eval {
                   select '(' || pymatrix(%s) || ',' || 
                      (select '[' 
                           || group_concat(pydict('modgen',''''||m.sagename||'''',
                                                  'resgen',subsel.resgen))
                           || ']' from smash_generators as subsel 
                           join module_generators as m on m.rowid = subsel.modgen 
                       where s.sdeg = subsel.sdeg and s.ideg = subsel.ideg and s.edeg = subsel.edeg) 
                   || ')'
                   from smash_boxes as s
                   %s
               }] ,
            """ % (what,cond) ) + "]"

        #print "res=", eval(res)
        mat = []
        for (matrix,basis) in eval(res):
            for row in matrix:
                mat.append( self._unpack_vector(row,basis) )
        return mat

    def cycles(self,s=None,n=None,i=None,e=None,nov=None):
        """
        Return cycles in a range of degrees
        """

        return self._getmatrix('cycles',s=s,n=n,i=i,e=e,nov=nov)
 
    def boundaries(self,s=None,n=None,i=None,e=None,nov=None):
        """
        Return boundaries in a range of degrees
        """

        return self._getmatrix('boundaries',s=s,n=n,i=i,e=e,nov=nov)
 
    def cycle(self,generator):
        """
        Return a representing cycle for a homology generator.
        """

        mat = self._getmatrix('homology',s=generator["s"],i=generator["i"],e=generator["e"])
        #print mat
        return mat[generator["enum"]]

    def g(self,s=None,n=None,num=None,id=None):
        """
        Construct a homology generator from id or sequence number
        """

        import sage
        (cond,nexp) = self._search_condition()

        if not id is None:
            cond = "where rowid = %d" % id
        else:
            cond = "where sdeg = %d and ideg = n2i(%d,%d) and basid = %d" % (s,s,n,num or 0)

        res = "[" + self.tcl.eval( """
               join [smashprod db eval {
                   select pydict('id',rowid,'s',s.sdeg,'i',s.ideg,'e',s.edeg,'n',%s,'num',s.basid,'enum',s.ebasid) from homology_generators as s %s
               }] ,
            """ % (nexp,cond) ) + "]"
        #print res
        return [sage.algebras.stgfr.FormalSumOfDict([(1,sage.algebras.stgfr.Generator(self,x))],parent=self._fsums) for x in eval(res)][0]
        

    def homology_class(self,cycle):
        """
        Express cycle in terms of homology generators
        """

        cmd = "unset -nocomplain hclvar; set hclvar {}"    
        for (coeff,smashgen) in cycle:
            (modgen,resgen) = smashgen
            #print "cf=",coeff," mgen=",modgen," rgen=",resgen
            cmd = cmd + ";lappend hclvar %d {%s} %d" % (coeff,modgen,resgen["id"]) 

            
        res = self.tcl.eval("""
                     %s
                     #puts hv=$hclvar
                     smashprod sage_homology_class $hclvar
                     """ % cmd)
        #print eval(res)
        return FormalSum([(cf,self.g(id=gid)) for (cf,gid) in eval(res)],parent=self._fsums)

    def _makeSageModule(self,region):

        from sage.algebras.steenrod_algebra import SteenrodAlgebra_generic, SteenrodAlgebra_mod_two

        self.tcl.eval("""
           if {[info commands sagemod] eq ""} {
               set smashdb [smashprod db]
               yacop::sagemodule create sagemod [%s::resolution config] $smashdb
           }
           """ % self._resip)

        # find the module generators that are already in the database
        modgens = list(self._module.basis(region))
        script = "set mgens {}\n"
        for g in modgens:
            degs = self._module.degrees(g)
            script += "lappend mgens {{%s} {s %d i %d e %d}}\n" % (g,degs["s"],degs["i"],degs["e"])
        todo = self.tcl.eval("%s ; sagemod addModuleGens $mgens" % script)
        if self._resolution._viewtype == "odd":
            A = SteenrodAlgebra_generic(self._resolution._prime)
        else:
            A = SteenrodAlgebra_mod_two()
        #print todo
        todo = eval(todo)
        #print todo
        resp = "unset -nocomplain df;unset -nocomplain ar;array set df {};array set ar {};"
        for g in todo["diff"]: 
            mg = modgens[g]
            resp += "set g {%s};set df($g) {};" %mg
            dg = self._module.diff(mg)
            for (cf,tg) in list(dg):
                resp += "lappend df($g) %d {%s}\n" % (cf,tg) 
        atodo = todo["actR"]
        for g in atodo:
            mg = modgens[g]
            resp += "set g {%s};set ar($g) {};" % mg
            for (yop,sop) in atodo[g]:
                resp += "lappend ar($g) {%s} {" % (yop)
                rs = self._module.actR(mg,sop)
                for (cf,tg) in list(rs):
                    resp += "%d {%s} " % (cf,tg)
                resp += "}\n" 
                    
        #print resp
        self.tcl.eval("""
              %s
              #parray df
              #parray ar
              sagemod addDiff [array get df]
              sagemod addActionR [array get ar]
              unset df
              unset ar
              #puts [sagemod db eval {select * from sagemod_ops}]
              #puts [smashprod db eval {select * from smash_fragments}]
           """ % resp)


class SteenrodAlgebraSmashableExampleMoore(SteenrodAlgebraSmashable):
    "Cohomology of the mod p moore space"

    def __init__(self,prime):
        if prime == 2:
            raise NotImplementedError, "current implementation is restricted to odd primes"
        self._prime = prime
        self._ring = GF(prime)

    def _repr_(self):
        return "mod %d Moore spectrum" % self._prime

    def _region_intersect(self,reg1,reg2):
        res = {}
        for l in ["s","i","e"]:
            key = l + "max"
            res[key] = min(reg1.get(key,+Infinity),reg2.get(key,+Infinity))
            key = l + "min"
            res[key] = max(reg1.get(key,-Infinity),reg2.get(key,-Infinity))
        return res

    def _region_contains(self,region,s,i,e):
        if region["smax"] < s:
            return False
        if region["emax"] < e:
            return False
        if region["imax"] < i:
            return False
        if region["smin"] > s:
            return False
        if region["emin"] > e:
            return False
        if region["imin"] > i:
            return False
        return True

    def basis(self,region=None):
        res = Set([])
        reg = {
            "smin" : -Infinity,
            "imin" : -Infinity,
            "emin" : -Infinity,
            "smax" : +Infinity,
            "imax" : +Infinity,
            "emax" : +Infinity,
            }
        if not region is None:
            reg = self._region_intersect(reg,region)
        if self._region_contains(reg,0,0,0):
            res = res.union(Set(['b']))
        if self._region_contains(reg,0,1,1):
            res = res.union(Set(['t']))
        return res

    def actL(self,basel,a):
        if a.is_unit():
            deg = 0
        else:
            deg = a.degree()
        if deg == 0:
            # hack: this gets the constant term in a:
            coeff = a._raw["milnor"][()]
            return FormalSum([(coeff,basel)],parent=self._ring)
        if deg != 1 or basel != 'b':
            return FormalSum([],parent=self._ring)
        try:
            # hack: this gets the coefficient of Q_0 in a:
            coeff = a._raw["milnor"][((0,), ())]
        except KeyError:
            return FormalSum([],parent=self._ring)
        return FormalSum([(coeff,'t')],parent=self._ring)

    def actR(self,basel,a):
        return self.actL(basel,a)

    def diff(self,basel):
        return FormalSum([],parent=self._ring)

    def bbox(self):
        res = {
            "smin" : 0,
            "imin" : 0,
            "emin" : 0,
            "smax" : 0,
            "imax" : 1,
            "emax" : 1,
            }
        return res

    def degrees(self,basel):
        return {
            'b': {
                "s":0, "i":0, "e":0, 
                },
            't' : {
                "s":0, "i":1, "e":1, 
                },
            }[basel]

    def _make_element_(self,string):
        return string


if False:        

    p=3

    execfile("local/lib/yacop/stgfr.py")

    fname = None
    #fname = ":memory:"

    #C=SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(p),filename=fname)
    #C=SteenrodAlgebraGroundFieldResolution(prime=2,profile='0 {3 2 1}',filename=":memory:")

    SteenrodAlgebraGroundFieldResolution=SteenrodAlgebraGroundFieldResolution
    C=SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(p),filename=fname)

    smax=33
    nmax=p*p*p*p*p
    C.extend(s=smax,n=nmax,quiet=False)
    print len(C.generators())

    if p != 2:
        M=Moore(p)
        print M.basis()
        print M.basis(M.bbox())
        print M.basis({"emax":0})
        print [(lambda x: (x,M.degrees(x)))(g) for g in M.basis()]

        X=SmashResolution(C,M,filename=":memory:")

        print latex(X)
        nmax=175
        X.extend(s=smax-5,n=nmax,quiet=False)
        X.extend(s=smax-2,n=nmax,quiet=False)
    X.gui()
    #C.gui()

    #X.tcl.eval("smashprod db trace puts")

    for nd in [10,11,12,27,82]:
        gens = X.homology(n=nd)
        print "dimension ", nd, ": gens ", gens
        for g in gens:
            s=g["s"]
            i=g["i"]
            sbas = X.smash_basis(s=s,i=i)
            print "  smash basis s=", s, " i=",i, " : ", sbas
            print "  cycles: ", X.cycles(s=s,i=i)
            print "  boundaries: ", X.boundaries(s=s,i=i)
            print "  representative of ", g, " = ", X.cycle(g)
            print "  clone of g: ", X.g(id=g["id"]), " = ", X.g(s=s,i=i,num=g["num"])
            for c in X.cycles(s=s,i=i):
                print "  generator from cycle", c, "=", X.homology_class(c)



