r"""
Steenrod algebra ground field resolutions

AUTHORS: - Christian Nassau (2009-03-27: version 1.0)

This package allows to compute a minimal resolution of a sub 
Hopf-algebra of the Steenrod algebra. 

INTRODUCTION:

A resolution is represented by an object of class ``SteenrodAlgebraGroundFieldResolution``. 
That name is clearly too long to be typed more than once, so we begin by defining a shorter alias name::

    sage: # define alias
    sage: SGFR = SteenrodAlgebraGroundFieldResolution
    sage: # construct a resolution object
    sage: C5 = SGFR(SteenrodAlgebra(5)) ; C5 # doctest note: random (data path may differ)
    GroundFieldResolution (p=5, profile="-1 full", filename="gfr-steenrod-5-E-1Rfull.db", viewtype=odd)

The resolution is stored in a single-file SQLite_ database. You can specify an arbitrary 
file name in the constructor. The special name ":memory:" creates an in-memory resolution that
is automatically forgotten when the sage session ends.

The profile of the subalgebra has an exterior algebra part and a reduced part. It is represented 
as "exterior-bitmask {list-of-integers}". For example, the algebra generated by `Q_0` and `Q_2` 
has profile ``5 {}`` and `A(2)=span(Q_0,P^1,P^p)` has ``7 {2 1}`` (at odd primes) or "0 {3 2 1}" (for p=2)::

    sage: C2 = SGFR(filename=":memory:", algebra = SteenrodAlgebra(2)) ; C2 # doctest note: random (might change)
    GroundFieldResolution (p=2, profile="0 full", filename=":memory:", viewtype=even)
    sage: tmf = SGFR(filename=":memory:", prime=2, profile="0 {3 2 1}") ; tmf # doctest note: random (might change)
    GroundFieldResolution (p=2, profile="0 {3 2 1}", filename=":memory:", viewtype=even)

The exterior part of the profile is also supported at the prime 2. 
For example, the following represents the `E_2` of the algebraic Novikov 
spectral sequence for `Ext(BP)` at the prime 2::

    sage: algNSS = SGFR(filename=":memory:", prime=2, profile="-1 full") ; algNSS # doctest note: random (might change)
    GroundFieldResolution (p=2, profile="-1 full", filename=":memory:", viewtype=odd)

For p=2 one can think of the Steenrod algebra as the reduced part of the "odd Steenrod algebra
at the prime 2", except for a rescaling of the degrees: 
a generator of `Ext^{s,i}` in the first sense belongs to `Ext^{s,2i}`
in the second. The *viewtype* parameter allows to switch between the two perspectives.
Viewtype "even" only works if the exterior part is empty.

COMPUTING A RESOLUTION:

Before you can use a resolution you must extend it to 
the degree that you need::

    sage: C2 = SGFR(filename=":memory:", algebra = SteenrodAlgebra(2))
    sage: C2.extend(s=20,n=40)
    sage: # count generators in topological dimension 31
    sage: len(C2.generators(n=31))
    22
    sage: # ask for generators outside of the computed range
    sage: len(C2.generators(n=45))
    0
    sage: C2.is_complete(s=3,n=45)
    False
    sage: # extend the resolution and ask again
    sage: C2.extend(s=20,n=50)
    sage: C2.is_complete(s=20,n=45)
    True
    sage: len(C2.generators(n=45))
    11

You can get a running progress report for the computation if you call *extend*
with the extra argument ``quiet=False``. You can always 
interrupt a running computation with Ctrl-C without harm.

Needless to say that you shouldn't run more than one computation on the 
same file at the same time.
But there's also no harm if you accidentally do: 
in this case the earlier process will terminate with 
the message "database has been hijacked" as soon as it tries to write
to the database.

USING THE RESOLUTION:

After the resolution has been computed you can use the *generators(...)* function
to find the generators in any given bidegree. Since the computed
resolution is minimal these provide a dual basis to `Ext_A(F_p,F_p)`::

    sage: # find a basis of Ext^{s,t}(F_2,F_2) for s=9, t-s=23
    sage: sorted(C2.generators(s=9,n=23))
    [g(9,23), g(9,23,1)]

As you can see Ext is 2-dimensional in that bidegree. The generators are displayed in the 
form *g(s,n,num)* where *s* is the homological degree, *n* the topological dimension 
and *num* the number of the generator in that bidegree. If the number is 0 it is omitted.
The `\text{\LaTeX}` representation is `g^{(s)}_{n}` or `g^{(s)}_{n,num}`.

Every generator has a unique id in the resolution, namely the *rowid* of 
the corresponding dataset. Although this *id* has no mathematical significance, 
it is occasionally useful to have it, for example if you are using an external tool 
to investigate the resolution datafile. You can get the *id* by looking 
at the "id" entry in the dictionary under the generator::
 
    sage: # find the id of the non-zero e_0 from Ext_A^{4,21}
    sage: e0 = C2.g(4,17)
    sage: e0["id"]                       # doctest note: random
    47
    sage: # create the generator from its id
    sage: e0 == C2.g(id = e0["id"])
    True

Every generator contains a reference to the resolution it belongs to. FormalSumOfDicts from 
different resolutions are never equal::

    sage: # create another resolution for the same algebra
    sage: C2b = SGFR(filename=":memory:", algebra = SteenrodAlgebra(2))
    sage: C2b.extend(s=5,n=10)
    sage: C2.g(2,3) == C2b.g(2,3)
    False
    sage: C2.g(id=1) == C2b.g(id=1)
    False 
    sage: # note that C2b sometimes uses different ids for corresponding generators: 
    sage: C2.g(id=6), C2b.g(id=6)        # doctest note: random
    (g(1,15), g(2,0))

The *diff(...)* method gives you the differential of a generator::

    sage: # differential of h_0h_2
    sage: C2.diff(C2.g(2,3))
    Sq(4)*g(1,0) + Sq(0,1)*g(1,1) + Sq(1)*g(1,3)

In other words, one has `d(h_0h_2) = Sq^4 h_0 + Sq(0,1) h_1 + Sq^1 h_2`.

The differentials are given as ``FormalSum`` objects::

    sage: type(C2.diff(C2.g(2,3)))
    <class 'sage.structure.formal_sum.FormalSum'>
    sage: C2.diff(C2.g(2,3)).parent()
    Abelian Group of all Formal Finite Sums over mod 2 Steenrod algebra
    sage: sorted(list(C2.diff(C2.g(2,3))))
    [(Sq(1), g(1,3)), (Sq(0,1), g(1,1)), (Sq(4), g(1,0))]

If you play around with this you will find that the differentials soon become forbiddingly
complicated. You can specify filter options in the *diff* routine to
get only parts of the data::

    sage: i = C2.g(7,23)
    sage: # get the full differential
    sage: C2.diff(i)
    (Sq(3,7)+Sq(6,6))*g(6,0) + (Sq(4,1,1)+Sq(11,1))*g(6,10) + (Sq(0,2,1)+Sq(7,2))*g(6,11) + (Sq(1,3)+Sq(4,2))*g(6,14) + Sq(5,1)*g(6,16) + (Sq(0,0,1)+Sq(1,2)+Sq(4,1))*g(6,17) + Sq(1,1)*g(6,20)
    sage: # extract only the P_t^s terms
    sage: C2.diff(i,ptsonly=True)
    Sq(0,0,1)*g(6,17)
    sage: # only fetch the piece that is attached to h_0^2d_0
    sage: C2.diff(i,target=C2.g(6,14))
    (Sq(1,3)+Sq(4,2))*g(6,14)
    sage: # filter by the degree of the Steenrod operations
    sage: C2.diff(C2.g(6,17),opdeg=3)
    Sq(0,1)*g(5,15,1)

Divisibility by the `h_k` is indicated by the presence of `Sq^{2^k}`
in the differential. You can search for these terms by combining *ptsonly* with *opdeg*::

    sage: # find h1-divisibility in dimension 17: look for all Sq^2 that originate there
    sage: sorted([(lambda x:(x,C2.diff(x,ptsonly=True,opdeg=2)))(g) for g in C2.generators(n=17)])
    [(g(3,17), Sq(2)*g(2,16)),
    (g(4,17), 0),
    (g(5,17), 0),
    (g(6,17), 0),
    (g(7,17), Sq(2)*g(6,16)),
    (g(8,17), Sq(2)*g(7,16)),
    (g(9,17), 0)]


For example, note that *g(7,17)* is divisible by `h_k` for all `k=0,1,2,3`::

    sage: C2.diff(C2.g(7,17),ptsonly=True)
    Sq(8)*g(6,10) + Sq(4)*g(6,14) + Sq(2)*g(6,16) + Sq(1)*g(6,17)

GRAPHICAL USER INTERFACE:

You can inspect the resolution with a graphical user interface using the *gui* method. 
This will open up a new window with an Ext chart in the usual format. 
You can also use this to create postscript charts or to access the underlying
sqlite database from a SQL console::

    sage: # start graphical user interface 
    sage: C2.gui()  # sage doctest note: not tested

You can also open the SQL console directly::

    sage: # open an SQL console
    sage: C2.sqlcon()  # sage doctest note: not tested

REFERENCES:

.. [Na] C. Nassau, "Ein neuer Algorithmus zur Untersuchung der Kohomologie der Steenrod-Algebra", 
	Logos Verlag Berlin 2002, ISBN 3-89722-881-5
.. [Null] The author's homepage `<http://www.nullhomotopie.de>`_. This should contain
          the sources for the Steenrod and Yacop libraries.
.. [SQLite] Project homepage `<http://www.sqlite.org>`_.


CLASS DOCUMENTATION:
"""

#*****************************************************************************
#       Copyright (C) 2009 Christian Nassau <nassau@nullhomotopie.de>
#  Distributed under the terms of the GNU General Public License (GPL)
#*****************************************************************************


class Yacop:
    """
    Interface to yet another cohomology program (YaCoP).
    """

    tcl = None
    "The main Tcl interpreter. Resolutions use a namespace of this as their interpreter."

    tcllibrary = None
    "Extra paths, added to auto_path when Tcl starts up"

    def __init__(self):
        pass

    class Slave:
        "A namespace in the main Yacop interpreter"

        def __init__(self):
            from Tkynter import Tcl, Tk
            if Yacop.tcl is None:
                Yacop.tcl = Tcl()
                xdirs = Yacop.tcllibrary
                if not xdirs is None:
                    if not  isinstance(xdirs,list):
                        xdirs = [ xdirs ]
                    [Yacop.tcl.eval("lappend auto_path {%s}" % path) for path in xdirs]
            Yacop.tcl.eval("""
                lappend auto_path $::env(SAGE_LOCAL)/lib
                package require yacop::sage 1.0
              """)
            self._slavename = Yacop.tcl.eval("yacop::sage::new-slave")
     
        def __del__(self):
            if not self._slavename is None:
                Yacop.tcl.eval("catch {namespace delete %s}" % self._slavename)
 
        def eval(self,script):
            "Execute a Tcl script in the slave's namespace"

            return Yacop.tcl.eval("namespace eval {%s} {%s}" % (self._slavename,script)) 

        def loadTk(self):
            "Load the Tk widget library into the slave's interpreter"

            self.eval("""
               namespace eval :: {package require Tk}
               if {![winfo exists .yacop]} {
                   set x {}
                   lappend x "I am the Yacop main window."
                   lappend x "You should not see me."
                   lappend x "If you close me, Yacop GUIs will no longer work."
                   label .yacop -text [join $x \\n]
                   pack .yacop -expand 1 -fill both
                   wm title . "Yacop main window"
                   proc ::yacop::progress {} {update}
               }
               wm withdraw .
             """ 
                    )
            
from sage.structure.sage_object import SageObject
from string import atoi
from sage.rings.all import GF
from sage.sets.set import Set
from sage.structure.formal_sum import FormalSum, FormalSums

class FormalSumOfDict(FormalSum):
    """
    A FormalSum variant with an extra __getitem__ method
    """

    def __getitem__(self, key):
        res = None
        for (cf,gen) in list(self):
            if not res is None:
                if gen[key] != res:
                    raise KeyError, "inconsistent entries for key %s" % key
            else:
                res = gen[key]
        if res is None:
            raise KeyError, "cannot determine '%s' because sum is zero" % key
        return res

class Generator(SageObject,dict):
    """
     A generator is a dictionary with some of the following keys
       :id:           unique identifier 
       :s:            homological degree
       :i:            internal degree
       :n:            topological dimension (n=i-s)
       :e:            exterior algebra degree
       :num:          bidegree sequence number, identifies generators with given s and n
       :enum:          bidegree sequence number, identifies generators with given s, n and e
       :resolution:   weak reference to the parent resolution       
    """

    def __init__(self,owner,dictval):
        from weakref import ref
        dict.__init__(self)
        self["resolution"] = ref(owner)
        for i in dictval:
            self[i] = dictval[i]

    def _repr_(self):
        num = self["num"]
        if num == 0:
            return "g(%d,%d)" % (self["s"], self["n"])
        return "g(%d,%d,%d)" % (self["s"], self["n"], num)

    def _latex_(self):
        num = self["num"]
        if num == 0:
            return "g^{(%d)}_{%d}" % (self["s"], self["n"])
        return "g^{(%d)}_{%d,%d}" % (self["s"], self["n"], num)

    def __hash__(self):
        return self["id"]

class SteenrodAlgebraGroundFieldResolution(SageObject):
    r"""
    Minimal resolution of `F_p` over a sub Hopf algebra of the Steenrod algebra.

    To construct a resolution you have to specify a meaningful subset of the arguments:
       :prime:        (requires that you also give the profile)
       :profile:      profile of the algebra to resolve, for example ``7 {2 1}`` for `A(2)`
       :algebra:      the algebra to resolve. specify *either* this or prime/profile
       :filename:     name of the database file to use. can be ":memory:" for an in-memory resolution
       :viewtype:     even/odd, p=2 only. this switches between `i` and `i/2`
       :quiet:        boolean to turn progress report on/off

    IMPLEMENTATION NOTES:

    Every resolution has its own Tcl interpreter with the Yacop package loaded into it.
    Packages are searched in the "local/lib" directory, although extra paths
    can be configured through the *Yacop.tcllibrary* variable.
    The interpreter is a slave of the main interpreter *Yacop.tcl*, which is common
    to all resolution related work.

    The interpreter is accessible from the outside through the "tcl" attribute.
    The resolution is represented in that interpreter by a "yacop::gfr" object
    with name "resolution". Its associated sqlite database can be accessed 
    through the "db" subcommand::

        sage: C = SteenrodAlgebraGroundFieldResolution(SteenrodAlgebra(2),filename=":memory:")
        sage: C.tcl.eval("resolution db eval {select value from yacop_cfg where key='prime'}")
        '2'
    """

    def __init__(self, algebra=None, prime=None, profile=None, 
                 filename=None, viewtype=None, quiet=True):
        
        self._filename = None
        self._algebra  = None
        self._prime    = None
        self._profile  = None
        self._profmode = 'auto'
        self._viewtype = 'odd'

        self.tcl = Yacop.Slave()

        if ((prime is None) and (not profile is None)) or ((not prime is None) and (profile is None)):
            raise ValueError, "prime and profile must be specified together"

        if (not prime is None) and (not algebra is None):
            raise ValueError, "cannot specify algebra and prime/profile at the same time"

        if not algebra is None:
            # translate algebra into prime/profile information
            # Why does this not work???: if isinstance(algebra, SteenrodAlgebra_generic):
            if hasattr(algebra, "prime"):
                prime = algebra.prime
                if 2 == prime:
                    profile = "0 full"
                    _viewtype = "even"
                else:
                    profile = "-1 full"
                    _viewtype = "odd"
            else:
                raise ValueError, "algebra %s not understood" % algebra
        
        if prime is None and filename is None:
            raise ValueError, "neither filename nor algebra given"

        if not prime is None:
            # test whether prime is supported
            self.tcl.eval("steenrod::prime %s inverse 1" % prime)
            self._prime = prime
            # test profile 
            self.tcl.eval("steenrod::algebra verify {%s}" % profile)
            self._profile = profile
            if not filename:
                filename = self.tcl.eval("""
                  set p %d
                  set a {%s}
                  set ae [lindex $a 0]
                  set ar [lindex $a 1]
                  set fnm "gfr-steenrod-$p-E${ae}R[join $ar -].db"
                  return $fnm
                 """ % (prime, profile))
                 
        if filename is None:
            filename = ":memory:"
        
        self._filename = filename
        if self._prime:
            self.tcl.eval("set p %d;set alg {%s}" % (self._prime, self._profile))

        self.tcl.eval("""
                 yacop::gfr create resolution {%s}
                 if {[info exists alg]} {
                    resolution algebra $p $alg
                 }
                 array set resinfo [resolution algebra]
                 if {$resinfo(prime) eq ""} {
                    error "cannot deduce the prime/algebra of this resolution"
                 }
                 set resinfo(viewtype) odd
                 if {$resinfo(prime) == 2 && 0 == [lindex $resinfo(algebra) 0]} {
                    set resinfo(viewtype) even
                 }
                 resolution viewtype $resinfo(viewtype)
            """ % self._filename)
        
        self._prime = atoi(self.tcl.eval("set resinfo(prime)"))
        self._profile = self.tcl.eval("set resinfo(algebra)")
        self._viewtype = self.tcl.eval("set resinfo(viewtype)")
        self._fsumsGF = FormalSums(GF(self._prime))
        if not viewtype is None:
            self._viewtype = viewtype
        self["viewtype"] = self._viewtype

        self["quiet"] = quiet

    def _repr_(self):
        return "GroundFieldResolution (p=%d, profile='%s', filename='%s')" % (self._prime, 
                                                                              self._profile, 
                                                                              self._filename)

    def _latex_(self):
        return "{\rm %s}" % self._repr_()

        
    def __getitem__(self, key):
        _getcode = {
            'filename': lambda x: self._filename,
            'algebra':  lambda x: self._algebra,
            'prime':    lambda x: self._prime,
            'profile':  lambda x: self._profile,
            'profmode': lambda x: self._profmode,
            'viewtype': lambda x: self._viewtype,
            'quiet':    lambda x: self._quiet,
            }
        if _getcode.has_key(key):
            return _getcode[key](x)
        raise ValueError, "unknown key \"%s\"" % key
    
    def _setprofmode(self, value):
        oklist = ["auto", "upper", "lower", "none"]
        if not (value in oklist):
            raise ValueError, "unknown profmode \"%s\", should be one of %s" % (value, oklist)
        self._profmode = value
 
    def _setviewtype(self, value):
        oklist = ["odd", "even"]
        if not (value in oklist):
            raise ValueError, "unknown viewtype \"%s\", should be one of %s" % (value, oklist)
        if value == "even" and self._prime != 2:
            raise ValueError, "viewtype \"even\" can only be used at the prime 2"
        self.tcl.eval("resolution viewtype %s" % value)
        self._viewtype = value

        # hmm, without the following line I get "NameError: global name 'sage' is not defined"
        import sage

        if self._viewtype == "even":
            self._fsums = sage.structure.formal_sum.FormalSums_generic(sage.algebras.steenrod_algebra.SteenrodAlgebra_mod_two())
        else:
            self._fsums = sage.structure.formal_sum.FormalSums_generic(sage.algebras.steenrod_algebra.SteenrodAlgebra_generic(self._prime))

    def _setquiet(self, value):
        if not isinstance(value, bool):
            raise ValueError, "quiet must be a bool"
        if value:
            self.tcl.eval("yacop::sectionizer quiet on")
        else:
            self.tcl.eval("yacop::sectionizer quiet off")
        self._quiet = value

    def __setitem__(self, key, value):
        _setcode = {
            'profmode': lambda x: self._setprofmode(value),
            'viewtype': lambda x: self._setviewtype(value),
            'quiet'   : lambda x: self._setquiet(value),
            }
        if _setcode.has_key(key):
            return _setcode[key]('dummy')
        raise ValueError, "key \"%s\" cannot be set" % key

    def gui(self):
        """
        Open a graphical user interface. This allows you to inspect the resolution, create
        postscript charts and to run SQL commands.
        """

        self.tcl.loadTk()
        self.tcl.eval("""
              set chv [yacop::chartgui [resolution db] {%s}]
              trace add variable [$chv forever] write "[list $chv destroy];#"
           """ % self._filename)
                        
    def sqlcon(self):
        """
        Open a SQL console to inspect the underlying database.
        """
        self.tcl.loadTk()
        self.tcl.eval("sqlconsole new [resolution db]")
        

    def is_complete(self, s, i=None, n=None):
        "Check whether the bidegree has been computed"
        if (i is None) and (n is None):
            raise ValueError, "either internal or topological degree must be given"
        if (not i is None) and (not n is None):
            raise ValueError, "both internal and topological degree given"
        if n :
            i = s+n
        if self._viewtype == "even":
            i = 2*i;
        res = self.tcl.eval("resolution isComplete %d %d" % (s,i))
        if eval(res):
            return True
        return False

    def extend(self, s, i=None, n=None, quiet=None):
        """
        Extend the resolution to the indicated bidegree. 
        """
        if (i is None) and (n is None):
            raise ValueError, "no bound for internal or topological degree given"
        if (not i is None) and (not n is None):
            raise ValueError, "cannot bound both internal and topological degree"
        bounds = " sdeg %d" % s
        if not i is None:
            bounds += " ideg %d" % i
        if not n is None:
            if self._viewtype == "even":
                bounds += " rtdeg %d" %n
            else:
                bounds += " tdeg %d" %n
        if  not quiet is None:
            self["quiet"] = quiet
        self.tcl.eval("""
              resolution profmode %s
              resolution extend-to {%s}
           """ % (self._profmode, bounds) )


    def generators(self, s=None, i=None, n=None, e=None, nov=None):
        """
        Search for generators with the indicated attributes.
        "nov" is just `s-e` (the name stands for *Novikov degree*).
        """
        import sage
        cond = ""
        if not s is None: cond += " and sdeg = %d" % s
        if not i is None: cond += " and ideg = %d" % i
        if not n is None: cond += " and ndeg = %d" % n
        if not e is None: cond += " and edeg = %d" % e
        if not nov is None: cond += " and (sdeg-edeg) = %d" % nov
        if cond != "": cond = "where " + cond[4:]
        res = "[" + self.tcl.eval( """
               join [resolution db eval {
                   select pydict('id',rowid,'s',sdeg,'i',ideg,'e',edeg,'n',ndeg,'num',basid) from chart_generators %s
               }] ,
            """ % cond ) + "]"
        return Set([sage.algebras.stgfr.FormalSumOfDict([(1,sage.algebras.stgfr.Generator(self,x))],parent=self._fsumsGF) for x in eval(res)])
        
    def g(self, s=None, n=None, num=None, id=None):
        """
        Create a generator from either *id* or bidegree *(s,n)* and sequence number *num*.
        """
        import sage
        if not id is None:
            if (not s is None) or (not n is None) or (not num is None):
                raise ValueError, "expected either id or s,n,num"
            res = self.tcl.eval( """
               resolution db onecolumn {
                   select pydict('id',rowid,'s',sdeg,'i',ideg,'e',edeg,'n',ndeg,'num',basid) from chart_generators where rowid = %d
               }
            """ % id )
        else:
            if (s is None) or (n is None):
                raise ValueError, "expected either id or s,n,num"
            if num is None:
                num = 0
            res = self.tcl.eval( """
               resolution db onecolumn {
                   select pydict('id',rowid,'s',sdeg,'i',ideg,'e',edeg,'n',ndeg,'num',basid) from chart_generators 
                   where sdeg = %d and ideg = %d and basid = %d
               }
            """ % (s,s+n,num) )
        if res == "":
            raise ValueError, "no such generator"
        return sage.algebras.stgfr.FormalSumOfDict([(1,sage.algebras.stgfr.Generator(self,eval(res)))],parent=self._fsumsGF)
    
    def diff(self, src, target=None, opdeg=None, ptsonly=False):
        """
        Get the differential of the generator *src*. The result is returned 
        as a list of pairs *(algebra element, generator)*.

        The differential can be filtered through these optional parameters:
           
             :target:      filter by target generator
             :opdeg:       filter by degree of the algebra element
             :ptsonly:     return only `P_t^s` and Bocksteins

        Note that you can search for a specific `P_t^s` or `Q_k` by combining *opdeg* and *ptsonly*.
        """

        from sage.algebras.steenrod_algebra import SteenrodAlgebra_mod_two, SteenrodAlgebra_generic

        cond = "srcgen = %d" % src["id"]
        if not (target is None):
            cond += " and targen = %d" % target["id"]
        if not (opdeg is None):
            if self._viewtype == "even":
                opdeg *= 2
            cond += " and opideg = %d" % opdeg
        funcname = "sagepoly_" + self._viewtype
        if ptsonly:
            funcname += "_PtsOnly"
        res = self.tcl.eval( """
               set result {}
               resolution db eval {
                   select %s(%d,group_concat(frag_decode(format,data),' '),targen) as spol from fragments 
                   where %s group by targen, format
               } {
                  if {$spol ne ""} {
                     lappend result $spol
                  }
               }
               return "\[[join $result ,]\]"
            """ % (funcname,self._prime,cond) )
        if self._viewtype == "odd":
            A = SteenrodAlgebra_generic(self._prime)
        else:
            A = SteenrodAlgebra_mod_two()
        return FormalSum(eval(res),parent=self._fsums)
        

if __name__ == "huhu":

    C2 = SteenrodAlgebraGroundFieldResolution(filename=":memory:", algebra = SteenrodAlgebra(2), quiet=True)
    #C2["quiet"] = False
    C2.extend(s=20,n=40)
    print C2.generators(s=4)
    print C2.diff(C2.g(4,23,0))
    print C2.diff(C2.g(4,23,0),ptsonly=True)
    print C2.diff(C2.g(7,17,0),ptsonly=True)


if False:
    SGFR = SteenrodAlgebraGroundFieldResolution
    smax = 20; nmax = 40
    Ares = SGFR(SteenrodAlgebra(2),filename=":memory:")
    Ares.extend(s=smax,n=nmax,quiet=True)
    a = Ares.g(9,23); b = Ares.g(9,23,1)

